/* ***** BEGIN LICENSE BLOCK *****
 *   Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Save File to.
 *
 * The Initial Developer of the Original Code is
 * Atte Kemppilä.
 * Portions created by the Initial Developer are Copyright (C) 2009
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 * 
 * ***** END LICENSE BLOCK ***** */


var EXPORTED_SYMBOLS =
[
    "CachingDataSource",
    "PreferenceDataSource",
    "TransformingDataSource"
];

const Cc = Components.classes;
const Ci = Components.interfaces;
const Cu = Components.utils;

Cu.import("resource://gre/modules/Services.jsm");
Cu.import("resource://gre/modules/XPCOMUtils.jsm");

Cu.import("resource://savefileto/common.jsm");


function PreferenceDataSource(prefName)
// extends ListenerManager
// implements nsIObserver
{
    ListenerManager.call(this);
    
    this.pref = Services.prefs.getBranch(prefName);
    this.pref.QueryInterface(Ci.nsIPrefBranch2);
    this.pref.addObserver("", this, true);
    
    this.listeners = [];
}

PreferenceDataSource.prototype =
{
    QueryInterface: XPCOMUtils.generateQI([
        Ci.nsIObserver,
        Ci.nsISupports,
        Ci.nsISupportsWeakReference]),
    
    get entries()
    {
        try
        {
            var json = this.pref.getComplexValue("", Ci.nsISupportsString).data;
            var entries = JSON.parse(json);
            return entries;
        }
        catch (e)
        {
            Cu.reportError(e);
            return [];
        }
    },
    
    set entries(entries)
    {
        if (! Array.isArray(entries))
            throw new Error("not an array: " + entries);
        
        var json = JSON.stringify(entries);
        var str = Cc["@mozilla.org/supports-string;1"].createInstance(Ci.nsISupportsString);
        str.data = json;
        this.pref.setComplexValue("", Ci.nsISupportsString, str);
    },
    
    dispose: function PrefDS_dispose()
    {
        this.pref.removeObserver("", this);
        
        delete this.pref;
        delete this._listeners;
    },
    
    // nsIObserver
    observe: function PrefDS_observe(subject, topic, data)
    {
        this.notifyListeners(this);
    }
}

extend(PreferenceDataSource, ListenerManager);


function DecoratingDataSource(dataSource)
{
    this.dataSource = dataSource;
}

DecoratingDataSource.prototype =
{
    get entries()
    {
        return this.dataSource.entries;
    },
    
    set entries(entries)
    {
        this.dataSource.entries = entries;
    },
    
    addListener: function DecoDS_addListener(callback)
    {
        return this.dataSource.addListener(callback);
    },
    
    removeListener: function DecoDS_removeListener(callback)
    {
        this.dataSource.removeListener(callback);
    },
    
    dispose: function DecoDS_dispose()
    {
        this.dataSource.dispose();
    }
}


function TransformingDataSource(dataSource, getTransformer, setTransformer)
// extends DecoratingDataSource
{
    DecoratingDataSource.call(this, dataSource);
    
    this.getTransformer = getTransformer;
    this.setTransformer = setTransformer;
}

TransformingDataSource.prototype =
{
    get entries()
    {
        var entries = this.dataSource.entries;
        entries = this._transform(this.getTransformer, entries);
        return entries;
    },
    
    set entries(value)
    {
        if (! Array.isArray(value))
            throw new Error("not an array: " + value);
        
        var entries = this._transform(this.setTransformer, value);
        this.dataSource.entries = entries;
    },
    
    _transform: function TransDS__transform(transformer, value)
    {
        return (transformer ? transformer.transform(value) : value);
    }
}

extend(TransformingDataSource, DecoratingDataSource);


function CachingDataSource(dataSource, writeThrough)
// extends DecoratingDataSource
{
    DecoratingDataSource.call(this, dataSource);
    
    this.writeThrough = writeThrough;
    
    var self = this;
    this.resetListener = dataSource.addListener(function() { self.reset(); });
}

CachingDataSource.prototype =
{
    get entries()
    {
        if (! this._entries)
            this._entries = this.dataSource.entries;
        
        return this._entries;
    },
    
    set entries(entries)
    {
        if (this.writeThrough)
        {
            this.dataSource.entries = entries;
            this.reset();
        }
        else
            this._entries = entries;
    },
    
    dispose: function CacheDS_dispose(shallow)
    {
        this.removeListener(this.resetListener);
        
        if (! shallow)
            this.dataSource.dispose();
        
        delete this._entries;
    },
    
    flush: function CacheDS_flush()
    {
        if ((! this.writeThrough) && this._entries)
            this.dataSource.entries = this._entries;
    },
    
    reset: function CacheDS_reset()
    {
        delete this._entries;
    }
}

extend(CachingDataSource, DecoratingDataSource);
